/********************************************************************\

  Name:         consume.c
  Created by:   Stefan Ritt

  Contents:     Buffer manager test program. Simple consumer connec-
                ting to a SYSTEM buffer and receiving some data.


  Revision history
  ------------------------------------------------------------------
  date        by    modification
  --------    ---   ------------------------------------------------
  11-NOV-93   SR    created
  25-JAN-95   SR    event buffer now dynamically 
                    (-> MS-DOS limited stack size)

\********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "midas.h"

INT    all_flag;
INT    hBufEvent;
char   event_buffer[MAX_EVENT_SIZE];

/*----- process_message --------------------------------------------*/

void process_message(HNDLE hBuf, HNDLE id, EVENT_HEADER *message)
{
  /* just print message text which comes after event header */
  printf("%s\n", message + 1);
}

/*----- process_event ----------------------------------------------*/

void process_event(HNDLE hBuf, HNDLE request_id, EVENT_HEADER *pheader)
{
static INT    ser, count, start_time, jam=0;

INT           stop_time;
INT           size, *pdata;
BUFFER_HEADER buffer_header;
double        rate;

  /* accumulate received event size */
  size = pheader->data_size;
  count += size;

  /* check if first and last word inside event is equal
     to size to check that nothing was overwritten... */

  if (!jam)
    {
    /* only test once */
    pdata   = (INT *) (pheader + 1);
    if (pdata[0] != size || pdata[size/4-1] != size)
      cm_msg(MDEBUG, "process_event", "data jam");
    jam = 1;
    }

  /* if only some events are requested, sleep a little bit
     to simulate a random event consumer */
  if (!all_flag)
    ss_sleep(10);

  /* if all events are requested, now check the serial number
     if no events are missing */
  if (all_flag && (INT) pheader->serial_number != ser + 1)
    cm_msg(MDEBUG, "process_event", "Serial number mismatch: Ser: %ld, OldSer: %ld, ID: %d, size: %ld\n",
      pheader->serial_number, ser, pheader->event_id, pheader->data_size);

  ser = pheader->serial_number;

  /* calculate rates each second */
  if (ss_millitime() - start_time > 1000)
    {
    stop_time = ss_millitime();
    rate = count / 1024.0 / 1024.0 / ((stop_time-start_time)/1000.0);

    /* get information about filling level of the buffer */
    bm_get_buffer_info(hBufEvent, &buffer_header);
    size = buffer_header.read_pointer - buffer_header.write_pointer;
    if (size <= 0)
      size += buffer_header.size;
    printf("Level: %4.1lf %%, ", 100-100.0*size/buffer_header.size);

    printf("Rate: %1.2lf MB/sec\n", rate);
    start_time = stop_time;
    count = 0;
    }
}

/*------------------------------------------------------------------*/

main()
{
INT    status, size;
char   host_name[30], str[30];
INT    read_cache_size;
INT    event_id, request_id;
DWORD  last_time;
BOOL   via_callback;

  /* get parameters */

  printf("ID of event to request: ");
  gets(str);
  event_id = atoi(str);

  printf("Host to connect: ");
  gets(host_name);

  printf("Get all events (0/1): ");
  gets(str);
  all_flag = atoi(str);

  printf("Read cache size: ");
  gets(str);
  read_cache_size = atol(str);

  printf("Receive via callback ([y]/n): ");
  gets(str);
  via_callback = str[0] != 'n';

  /* connect to experiment */
  status = cm_connect_experiment(host_name, "", 
             all_flag ? "Power Consumer" : "Consumer", NULL);
  if (status == CM_WRONG_PASSWORD)
    {
    printf("Wrong password\n");
    return 0;
    }
  if (status != CM_SUCCESS)
    {
    printf("Cannot connect to host '%s'.\n", host_name);
    return 0;
    }

#ifdef _DEBUG
  cm_set_watchdog_params(TRUE, 0);
#endif

  /* open the "system" buffer, 1M size */
  bm_open_buffer("SYSTEM", EVENT_BUFFER_SIZE, &hBufEvent);

  /* set the buffer cache size if requested */
  bm_set_cache_size(hBufEvent, read_cache_size, 0);

  /* place a request for a specific event id */
  bm_request_event(hBufEvent, (WORD) event_id, TRIGGER_ALL,
                   all_flag ? GET_ALL : GET_SOME, &request_id, 
                   via_callback ? process_event : NULL);
  
  /* place a request for system messages */
  cm_msg_register(process_message);

  last_time = 0;

  do
    {
    if (via_callback)
      status = cm_yield(1000);
    else
      {
      /* receive event "manually" and call receive_event */
      size = sizeof(event_buffer);
      status = bm_receive_event(hBufEvent, event_buffer, &size, ASYNC);
      if (status == BM_SUCCESS)
        process_event(hBufEvent, request_id, (EVENT_HEADER *) event_buffer);

      /* call yield once every second */
      if (ss_millitime() - last_time > 1000)
        {
        last_time = ss_millitime();
        status = cm_yield(0);
        }
      }

    } while (status != RPC_SHUTDOWN && status != SS_ABORT);

  cm_disconnect_experiment();

  return 1;
}
